# This file implements a command line script to communicate with a smcarcard
#using pysc
#
# Copyright (C) 2013 Omar Choudary (omar.choudary@cl.cam.ac.uk)
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# - Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
#
# - Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import array
import serial
import sys
import shlex, subprocess
import time
import argparse # you need Python v2.7 or later
from tlv import T
import emv_commands
from atcmds import *
from scdtrace import *
from smartcard.Exceptions import NoCardException
from smartcard.System import readers
from smartcard.util import toHexString
from smartcard.CardType import AnyCardType
from smartcard.CardRequest import CardRequest
from smartcard.CardConnectionObserver import ConsoleCardConnectionObserver
from smartcard.Exceptions import CardRequestTimeoutException

def sc_getatr():
  """
  Returns the ATR from the card
  """

  # Get ATR
  for reader in readers():
    try:
        connection = reader.createConnection()
        connection.connect()
        print reader
        atr = connection.getATR()
        print toHexString(atr)
        return atr
    except NoCardException:
        print reader
        print 'no card inserted'

  return True

def sc_terminal(fid = sys.stdin):
  """
  Interactive terminal using pyscard library

  Args:
    fid is the file descriptor for the file containing the sequence of commands.

  Returns: True if ended correctly, False otherwise
  """

  # request any card type
  cardtype = AnyCardType()

  try:
    # request card insertion
    print 'insert a card (SIM card if possible) within 10s'
    cardrequest = CardRequest(timeout=10, cardType=cardtype)
    cardservice = cardrequest.waitforcard()

    # attach the console tracer
    observer = ConsoleCardConnectionObserver()
    cardservice.connection.addObserver(observer)

    # connect to the card and perform a few transmits
    cardservice.connection.connect()

    while True:
      line = fid.readline()
      if line.find('0000000000') == 0:
        fid.close()
        print 'Response: ', line
        return True
      else:
        cmd = line.rstrip('\n')
        print 'Sending CAPDU: ', cmd
        cmdhex = cmd.decode('hex')
        cmdarray = array.array('B', cmdhex)
        cmdlist = cmdarray.tolist()
        response, sw1, sw2 = cardservice.connection.transmit(cmdlist)

  except CardRequestTimeoutException:
      print 'time-out: no card inserted during last 10s'

  except:
      import sys
      print sys.exc_info()[1]

  return True

def scd_sc_link(port):
  """
  Connects a smartcard from a USB smartcard reader (via pyscard) to a
  GSM device using the SCD as a middle-device that forward commands and
  responses between the SIM card in the USB smartcard reader and the GSM device.

  Args:
    port is the port used to connect via seria/virtual serial to the SCD

  Returns: True if ended correctly, False otherwise.
  """

  #Get ATR from card
  atr = sc_getatr()
  atr = toHexString(atr)
  atr = atr.replace(" ", "")
  #ignore first byte as we don't use that
  atr = atr[2:]
  print 'ATR: ' + atr

  # request any card type
  cardtype = AnyCardType()

  try:
    # request card insertion
    print 'insert a card (SIM card if possible) within 10s'
    cardrequest = CardRequest(timeout=10, cardType=cardtype)
    cardservice = cardrequest.waitforcard()

    # attach the console tracer
    observer = ConsoleCardConnectionObserver()
    cardservice.connection.addObserver(observer)

    # connect to the card and perform a few transmits
    cardservice.connection.connect()

    # Initialize scd-terminal connection
    ser = serial.Serial(port)
    ser.write(AT_CMD.AT_CTUSB)
    ser.flush()
    line = ser.readline()
    if line.find('AT OK') < 0:
      print 'Error initialising scd-terminal'
      ser.close()
      return False

    #External while to allow for resets from GSM device (as I've seen from GSM dongle)
    while True:
      # Send ATR to terminal and get first command. Ignore first byte of ATR
      print time.time()
      print 'Sending data: ', atr
      cmd = 'AT+UDATA=' + atr + "\r\n"
      ser.write(cmd)
      ser.flush()
      line = ser.readline()
      print time.time()
      print 'Data from Terminal: ', line
      if line.find('AT OK') >= 0:
        return True
      elif line.find('AT BAD') >= 0:
        return False
      elif line.find('RESET') >= 0:
        print 'Got RESET from SCD'
        continue;

      #Internal while: forward commands and responses
      while True:
        # Send command to card and get response
        cmd = line.rstrip('\n').rstrip('\r')
        print time.time()
        print 'Sending CAPDU to ICC: ', cmd
        print emv_commands.command_name(cmd[0:4].lower())
        cmdhex = cmd.decode('hex')
        cmdarray = array.array('B', cmdhex)
        cmdlist = cmdarray.tolist()
        response, sw1, sw2 = cardservice.connection.transmit(cmdlist)
        rsp = toHexString(response)
        rsp = rsp.replace(" ", "")
        sw1 = hex(sw1)[2:].zfill(2)
        sw2 = hex(sw2)[2:].zfill(2)
        print 'Got response: ', rsp + sw1 + sw2
        status = sw1 + sw2
        print emv_commands.response_name(status.lower())
        if len(rsp) > 0:
          try:
            t = T(rsp)
            print t.dump()
          except Exception as (e):
            print e
            pass


        #Forward response to terminal via SCD and get next command
        if len(rsp) > 0:
          #Need to prepend INS byte, as the pyscard doesn't provide that (seems bad to me)
          rsp = cmd[2:4] + rsp
        rstr = rsp + sw1 + sw2
        print time.time()
        print 'Sending RAPDU to SCD: ', rstr
        cmd = 'AT+UDATA=' + rstr + "\r\n"
        ser.write(cmd)
        ser.flush()
        line = ser.readline()
        print time.time()
        print 'Data from Terminal: ', line
        if line.find('AT OK') >= 0:
          return True
        elif line.find('AT BAD') >= 0:
          return False
        elif line.find('RESET') >= 0:
          break;

  except CardRequestTimeoutException:
    print 'time-out: no card inserted during last 10s'

  except:
    import sys
    print sys.exc_info()[1]

  # Use this for ending
  #ser.write(AT_CMD.AT_CCEND)
  #ser.flush()
  #line = ser.readline()
  #ser.close();
  #if line.find('AT OK') >= 0:
  #  return True
  #else:
  #  return False


def main():
  """Command line for pyscard communication"""

  parser = argparse.ArgumentParser(description='SC Command Line')
  parser.add_argument(
      'port',
      help='the name of the serial port to communicate to the SCD, e.g. /dev/ttyACM0')
  parser.add_argument(
      '--scterminal',
      nargs = '?',
      type = argparse.FileType('r'),
      const = sys.stdin,
      default = False, metavar = 'filename',
      help='Act as a terminal and send the commands (CAPDUs) from the given filename (default stdin).\
          The CAPDUs must be complete commands (header + data) with no spaces in between; one per line.\n\
          The file (or sequence of commands) should end with the string "0000000000".\n\
          Example:                         \n\
          00A4040007A0000000048002  (SELECT) \n\
          80A80000028300            (GET PROCESSING OPTS)\n\
          .... \n\
          0000000000'),
  parser.add_argument(
      '--scdlink',
      action = 'store_true',
      help='Run a link between SCD and USB smartcard reader'),
  parser.add_argument('-v',
      '--verbose',
      action = 'store_true',
      help='be more verbose')
  args = parser.parse_args()

  if args.scterminal != False:
    try:
      print "Starting user terminal...\n"
      result = sc_terminal(args.scterminal)
      if result == True:
        print "All done"
      else:
        print "Some error ocurred during communication, check log"
    except:
      print "Error occurred"
      raise
  elif args.scdlink != False:
    try:
      print "Starting forwarding between SIM and SCD, follow SCD screen..."
      while True:
        print "#############################################################################################################################"
        result = scd_sc_link(args.port)
      if result == True:
        print "All done"
      else:
        print "Some error ocurred during communication, check log"
    except:
      print "Error occurred"
      raise

if __name__ == '__main__':
  main()

